#! /usr/bin/env python3
# -*- coding: utf-8 -*-


import arcade
import os

import math
import pymunk

from create_level import create_floor, create_platform, create_level_1

import timeit
from PIL import Image


# Constants
SCREEN_WIDTH = 1200
SCREEN_HEIGHT = 720
SCREEN_TITLE = "Platformer"

# Constants used to scale our sprites from their original size
CHARACTER_SCALING = 1
TILE_SCALING = 0.5
COIN_SCALING = 0.5
SPRITE_PIXEL_SIZE = 128
GRID_PIXEL_SIZE = SPRITE_PIXEL_SIZE * TILE_SCALING

# Movement speed of player, in pixels per frame
PLAYER_MOVEMENT_SPEED = 8
GRAVITY = 1.4
PLAYER_JUMP_SPEED = 32

# How many pixels to keep as a minimum margin between the character
# and the edge of the screen.
LEFT_VIEWPORT_MARGIN = 200
RIGHT_VIEWPORT_MARGIN = 200
BOTTOM_VIEWPORT_MARGIN = 150
TOP_VIEWPORT_MARGIN = 100

PLAYER_START_X = 256
PLAYER_START_Y = 256


class PhysicsSprite(arcade.Sprite):
    def __init__(self, pymunk_shape, filename):
        super().__init__(
            filename,
            center_x=pymunk_shape.body.position.x,
            center_y=pymunk_shape.body.position.y,
        )
        self.pymunk_shape = pymunk_shape


class CircleSprite(PhysicsSprite):
    def __init__(self, pymunk_shape, filename):
        super().__init__(pymunk_shape, filename)
        self.width = pymunk_shape.radius * 2
        self.height = pymunk_shape.radius * 2


class BoxSprite(PhysicsSprite):
    def __init__(self, pymunk_shape, filename, width, height):
        super().__init__(pymunk_shape, filename)
        self.width = width
        self.height = height


def make_sprite(mass, image, position, space):

    width, height = Image.open(image).size
    mass = mass
    moment = pymunk.moment_for_box(mass, (width, height))
    body = pymunk.Body(mass, moment)
    body.position = pymunk.Vec2d(position)
    shape = pymunk.Poly.create_box(body, (width, height))
    shape.friction = 0.3
    space.add(body, shape)
    sprite = BoxSprite(shape, image, width=width, height=height)
    return sprite


class MyGame(arcade.Window):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer for the game
        """

        # Call the parent class and set up the window
        super().__init__(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_TITLE)

        # Set the path to start with this program
        file_path = os.path.dirname(os.path.abspath(__file__))
        os.chdir(file_path)

        # These are 'lists' that keep track of our sprites. Each sprite should
        # go into a list.
        self.coin_list = None
        self.wall_list = None
        self.background_list = None
        self.ladder_list = None
        self.player_list = None

        # Separate variable that holds the player sprite
        self.player_sprite = None

        # Our 'physics' engine
        self.physics_engine = None

        # Used to keep track of our scrolling
        self.view_bottom = 0
        self.view_left = 0

        # Keep track of the score
        self.score = 0

        # Load sounds
        self.collect_coin_sound = arcade.load_sound("sounds/coin1.wav")
        self.jump_sound = arcade.load_sound("sounds/jump1.wav")
        self.game_over = arcade.load_sound("sounds/gameover1.wav")

        # _____________________________
        # self.parabolic_points = []
        # self.dx = self.dy = 0

        #
        self.space = pymunk.Space()
        # self.space.gravity = (0.0, -900.0)
        self.space.gravity = (0.0, -100.0)

        # Lists of sprites
        self.dynamic_sprite_list = arcade.SpriteList()  # !!
        self.static_sprite_list = arcade.SpriteList()  # !!

        self.sprite_list = arcade.SpriteList()  # !

        self.background = arcade.Sprite("images/background.png")
        self.background.left = self.background.bottom = 0

        self.static_lines = []

        # self.shape_being_dragged = None
        # self.last_mouse_position = 0, 0

        self.draw_time = 0
        self.processing_time = 0

        create_level_1(self.space, self.static_sprite_list, self.dynamic_sprite_list)

        # floor_height = 80
        floor_height = 80
        far_right = 500

        body = pymunk.Body(body_type=pymunk.Body.STATIC)
        # shape = pymunk.Segment(body, [0, floor_height], [SCREEN_WIDTH, floor_height], 0.0)
        shape = pymunk.Segment(body, [0, 80], [SCREEN_WIDTH * 3, 80], 0.0)
        shape.friction = 10
        self.space.add(shape)
        self.static_lines.append(shape)

        # 开始叠箱子--请用for循环优化下面的代码
        box = make_sprite(43, "images/boxCrate_double.png", (1000, 130), self.space)
        self.sprite_list.append(box)
        box = make_sprite(43, "images/boxCrate_double.png", (1000, 230), self.space)
        self.sprite_list.append(box)
        box = make_sprite(43, "images/boxCrate_double.png", (1000, 330), self.space)
        self.sprite_list.append(box)
        box = make_sprite(43, "images/boxCrate_double.png", (1000, 430), self.space)
        self.sprite_list.append(box)
        box = make_sprite(43, "images/boxCrate_double.png", (1000, 530), self.space)
        self.sprite_list.append(box)

        sprite = make_sprite(33, "images/pig.png", (1000, 620), self.space)
        self.sprite_list.append(sprite)

        sprite = make_sprite(33, "images/platform.png", (160, 160), self.space)
        self.sprite_list.append(sprite)

        metalsprite = make_sprite(33, "images/metal_128.png", (750, 620), self.space)
        self.sprite_list.append(metalsprite)

        metalsprite = make_sprite(33, "images/metal_128.png", (750, 800), self.space)
        self.sprite_list.append(metalsprite)

        metalsprite = make_sprite(500, "images/metal_256.png", (750, 1200), self.space)
        self.sprite_list.append(metalsprite)

        self.shoot_position = pymunk.Vec2d(PLAYER_START_X, PLAYER_START_Y)
        self.physic_player = make_sprite(
            133, "images/player_stand.png", self.shoot_position, self.space
        )
        self.sprite_list.append(self.physic_player)

        self.virtual_player = arcade.Sprite("images/player_stand.png")
        self.virtual_player.center_x = self.shoot_position[0]
        self.virtual_player.center_y = self.shoot_position[1]

        self.reset_shoot = True
        self.shape_being_dragged = None

        print("end of init : sprite_list = \n")
        [print(sprite) for sprite in self.sprite_list]

    def setup(self):
        """ Set up the game here. Call this function to restart the game. """

        # Used to keep track of our scrolling
        self.view_bottom = 0
        self.view_left = 0

        # Keep track of the score
        self.score = 0

        # Create the Sprite lists
        self.player_list = arcade.SpriteList()
        self.background_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList()
        self.coin_list = arcade.SpriteList()

        # Set up the player, specifically placing it at these coordinates.
        self.virtual_player = self.player_sprite = arcade.Sprite(
            "images/player_1/player_stand.png", CHARACTER_SCALING
        )
        # self.player_sprite = arcade.Sprite("../../../arcade/examples/platform_tutorial/images/player_1/player_stand.png")
        self.player_sprite.center_x = PLAYER_START_X
        self.player_sprite.center_y = PLAYER_START_Y
        self.player_list.append(self.virtual_player)

        self.shoot_position = pymunk.Vec2d(PLAYER_START_X, PLAYER_START_Y)
        self.physic_player = make_sprite(
            133, "images/player_stand.png", self.shoot_position, self.space
        )
        self.sprite_list.append(self.physic_player)

        self.virtual_player = arcade.Sprite("images/player_stand.png")
        self.virtual_player.center_x = self.shoot_position[0]
        self.virtual_player.center_y = self.shoot_position[1]

        # --- Load in a map from the tiled editor ---

        # Name of the layer in the file that has our platforms/walls
        platforms_layer_name = "Platforms"
        moving_platforms_layer_name = "Moving Platforms"

        # Name of the layer that has items for pick-up
        coins_layer_name = "Coins"

        map_name = f"planet.tmx"

        # Read in the tiled map
        my_map = arcade.tilemap.read_tmx(map_name)

        # Calculate the right edge of the my_map in pixels
        self.end_of_map = my_map.map_size.width * GRID_PIXEL_SIZE

        print("platforms_layer_name  :  ", platforms_layer_name)

        # -- Platforms
        self.wall_list = arcade.tilemap.process_layer(
            my_map, platforms_layer_name, TILE_SCALING
        )
        # debug
        for wall in self.wall_list:
            # print("\n\n\n wall : ", dir(wall))
            print(
                "\n\n\n wall : ",
                wall.right,
                "  ",
                wall.left,
                "  ",
                wall.position,
                "  ",
                wall.center_x,
                "  ",
                wall.center_y,
                "     ",
                wall.texture.image,
                "     ",
                wall.texture.name,
                "      ",
                wall.texture.texture,
            )
            print("\n\n\n")

        # -- Moving Platforms
        moving_platforms_list = arcade.tilemap.process_layer(
            my_map, moving_platforms_layer_name, TILE_SCALING
        )
        for sprite in moving_platforms_list:
            self.wall_list.append(sprite)

        # -- Background objects
        # self.background_list = arcade.tilemap.process_layer(my_map, "Background", TILE_SCALING)

        # -- Background objects
        self.ladder_list = arcade.tilemap.process_layer(my_map, "Ladders", TILE_SCALING)

        # -- Coins
        self.coin_list = arcade.tilemap.process_layer(
            my_map, coins_layer_name, TILE_SCALING
        )

        # --- Other stuff
        # Set the background color
        if my_map.background_color:
            arcade.set_background_color(my_map.background_color)

        # Create the 'physics engine'
        self.physics_engine = arcade.PhysicsEnginePlatformer(
            self.player_sprite,
            self.wall_list,
            gravity_constant=GRAVITY,
            ladders=self.ladder_list,
        )

    def on_draw(self):
        """ Render the screen. """

        # Clear the screen to the background color
        arcade.start_render()

        #
        draw_start_time = timeit.default_timer()

        # Draw our sprites
        self.wall_list.draw()
        self.background_list.draw()
        self.ladder_list.draw()
        self.coin_list.draw()
        self.player_list.draw()

        # Draw our score on the screen, scrolling it with the viewport
        score_text = f"Score: {self.score}"
        arcade.draw_text(
            score_text,
            10 + self.view_left,
            10 + self.view_bottom,
            arcade.csscolor.BLACK,
            18,
        )

        self.background.draw()

        self.sprite_list.draw()

        for line in self.static_lines:
            body = line.body
            pv1 = body.position + line.a.rotated(body.angle)
            pv2 = body.position + line.b.rotated(body.angle)
            arcade.draw_line(pv1.x, pv1.y, pv2.x, pv2.y, arcade.color.WHITE, 2)

        if self.reset_shoot:
            self.virtual_player.draw()

        x1, y1 = self.shoot_position
        x2, y2 = self.virtual_player.position
        dd = (x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)

        if dd > 2 and self.reset_shoot:
            arcade.draw_line(x1, y1, x2, y2, arcade.color.ORANGE, 4)

        print(len(self.parabolic_points), self.reset_shoot, self.dx, self.dy)
        if self.reset_shoot:
            for point in self.parabolic_points:
                arcade.draw_circle_outline(
                    point[0], point[1], 5, arcade.color.WISTERIA, 3
                )

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key == arcade.key.UP or key == arcade.key.W:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = PLAYER_MOVEMENT_SPEED

                # self.physic_player.pymunk_shape.body.position.y = PLAYER_MOVEMENT_SPEED
                # self.physic_player.pymunk_shape.body.force = -self.dx*20000, -self.dy*20000

            elif self.physics_engine.can_jump():
                self.player_sprite.change_y = PLAYER_JUMP_SPEED
                arcade.play_sound(self.jump_sound)

                # self.physic_player.pymunk_shape.body.position.y += PLAYER_MOVEMENT_SPEED
                # self.physic_player.pymunk_shape.body.force = -self.dx*20000, -self.dy*20000

        elif key == arcade.key.DOWN or key == arcade.key.S:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = -PLAYER_MOVEMENT_SPEED

                # self.physic_player.pymunk_shape.body.position.y += PLAYER_MOVEMENT_SPEED
                # self.physic_player.pymunk_shape.body.force = -self.dx*20000, -self.dy*20000

        elif key == arcade.key.LEFT or key == arcade.key.A:
            self.player_sprite.change_x = -PLAYER_MOVEMENT_SPEED

            # self.physic_player.pymunk_shape.body.position.x -= PLAYER_MOVEMENT_SPEED
            # self.physic_player.pymunk_shape.body.force = -self.dx*20000, -self.dy*20000

        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.player_sprite.change_x = PLAYER_MOVEMENT_SPEED

            # self.physic_player.pymunk_shape.body.position.x += PLAYER_MOVEMENT_SPEED
            # self.physic_player.pymunk_shape.body.force = -self.dx*20000, -self.dy*20000

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key == arcade.key.UP or key == arcade.key.W:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = 0

        elif key == arcade.key.DOWN or key == arcade.key.S:
            if self.physics_engine.is_on_ladder():
                self.player_sprite.change_y = 0

        elif key == arcade.key.LEFT or key == arcade.key.A:
            self.player_sprite.change_x = 0

        elif key == arcade.key.RIGHT or key == arcade.key.D:
            self.player_sprite.change_x = 0

        def on_mouse_press(self, x, y, button, modifiers):
            pass

            # if button == arcade.MOUSE_BUTTON_LEFT :
            #    self.shape_being_dragged = self.virtual_bird

    def on_mouse_release(self, x, y, button, modifiers):
        pass

    def on_mouse_motion(self, x, y, dx, dy):
        pass

    def update(self, delta_time):
        """ Movement and game logic """

        # Call update on all sprites (The sprites don't do much in this
        # example though.)
        self.physics_engine.update()

        # Update animations
        self.coin_list.update_animation(delta_time)
        self.background_list.update_animation(delta_time)

        # Update walls, used with moving platforms
        self.wall_list.update()

        # See if the wall hit a boundary and needs to reverse direction.
        for wall in self.wall_list:

            if (
                wall.boundary_right
                and wall.right > wall.boundary_right
                and wall.change_x > 0
            ):
                wall.change_x *= -1
            if (
                wall.boundary_left
                and wall.left < wall.boundary_left
                and wall.change_x < 0
            ):
                wall.change_x *= -1
            if wall.boundary_top and wall.top > wall.boundary_top and wall.change_y > 0:
                wall.change_y *= -1
            if (
                wall.boundary_bottom
                and wall.bottom < wall.boundary_bottom
                and wall.change_y < 0
            ):
                wall.change_y *= -1

        # See if we hit any coins
        coin_hit_list = arcade.check_for_collision_with_list(
            self.player_sprite, self.coin_list
        )

        # Loop through each coin we hit (if any) and remove it
        for coin in coin_hit_list:

            # Figure out how many points this coin is worth
            if "Points" not in coin.properties:
                print("Warning, collected a coing without a Points property.")
            else:
                points = int(coin.properties["Points"])
                self.score += points

            # Remove the coin
            coin.remove_from_sprite_lists()
            arcade.play_sound(self.collect_coin_sound)

        # Track if we need to change the viewport
        changed_viewport = False

        start_time = timeit.default_timer()
        self.virtual_player.update()

        for sprite in self.sprite_list:
            if sprite.pymunk_shape.body.position.y < 0:

                self.space.remove(sprite.pymunk_shape, sprite.pymunk_shape.body)

                sprite.kill()

        self.space.step(1 / 60.0)

        for sprite in self.sprite_list:
            sprite.center_x = sprite.pymunk_shape.body.position.x
            sprite.center_y = sprite.pymunk_shape.body.position.y
            sprite.angle = math.degrees(sprite.pymunk_shape.body.angle)

        self.processing_time = timeit.default_timer() - start_time

        # --- Manage Scrolling ---

        # Scroll left
        left_boundary = self.view_left + LEFT_VIEWPORT_MARGIN
        if self.player_sprite.left < left_boundary:
            self.view_left -= left_boundary - self.player_sprite.left
            changed_viewport = True

        # Scroll right
        right_boundary = self.view_left + SCREEN_WIDTH - RIGHT_VIEWPORT_MARGIN
        if self.player_sprite.right > right_boundary:
            self.view_left += self.player_sprite.right - right_boundary
            changed_viewport = True

        # Scroll up
        top_boundary = self.view_bottom + SCREEN_HEIGHT - TOP_VIEWPORT_MARGIN
        if self.player_sprite.top > top_boundary:
            self.view_bottom += self.player_sprite.top - top_boundary
            changed_viewport = True

        # Scroll down
        bottom_boundary = self.view_bottom + BOTTOM_VIEWPORT_MARGIN
        if self.player_sprite.bottom < bottom_boundary:
            self.view_bottom -= bottom_boundary - self.player_sprite.bottom
            changed_viewport = True

        if changed_viewport:
            # Only scroll to integers. Otherwise we end up with pixels that
            # don't line up on the screen
            self.view_bottom = int(self.view_bottom)
            self.view_left = int(self.view_left)

            # Do the scrolling
            arcade.set_viewport(
                self.view_left,
                SCREEN_WIDTH + self.view_left,
                self.view_bottom,
                SCREEN_HEIGHT + self.view_bottom,
            )


def main():
    """ Main method """
    window = MyGame()
    window.setup()
    arcade.run()


if __name__ == "__main__":
    main()
